unit NGraphics;

{$IMPORTEDDATA OFF}
{$MINENUMSIZE 4}

interface

uses Windows, WinAdd;

type
  TTextAlign = (taLeft=TA_LEFT, taRight=TA_RIGHT, taCenter=TA_CENTER);
  TLogicalFont = (lfStatus);
  TFontQuality = (
    fqDefault=DEFAULT_QUALITY,
    fqNonAntialiased=NONANTIALIASED_QUALITY,
    fqAntialiased=ANTIALIASED_QUALITY,
    fqClearType=CLEARTYPE_QUALITY
  );

  TFont = class
    Handle: HFONT;
    Name: WideString;
    Size: Integer;
    Bold: Boolean;
    Italic: Boolean;
    Quality: TFontQuality;
    constructor Create(const Name: WideString; Size: Integer;
        Bold, Italic: Boolean; Quality: TFontQuality=fqDefault); overload;
    constructor Create(LogicalFont: TLogicalFont); overload;
    destructor Destroy; override;
  end;

  TTextMetrics = record
    Height: Integer;
    InternalLeading: Integer;
  end;

  TCanvas = class
  private
    Brush: HBRUSH;
    OldFont: HFONT;
    OldPen: HPEN;
    procedure UpdateBrush(Color: Cardinal);
    procedure UpdatePen(Color: Cardinal);
  public
    Handle: HDC;
    procedure BeginDraw;
    procedure EndDraw;
    procedure SetTextColor(Color: Cardinal);
    procedure SetTextAlign(Align: TTextAlign);
    procedure SetFont(const Font: TFont);
    procedure DrawText(X, Y: Integer; const S: string); overload;
    procedure DrawText(X, Y: Integer; const S: WideString); overload;
    procedure DrawRect(X, Y, W, H: Integer; Color: Cardinal);
    procedure FillRect(X, Y, W, H: Integer; Color: Cardinal);
    procedure Line(X1, Y1, X2, Y2: Integer; Color: Cardinal);
    function TextMetrics: TTextMetrics;
  end;

  TMemCanvas = class(TCanvas)
  private
    Wnd: HWND;
    Bpp: Integer;
    Bitmap: HBITMAP;
    Width: Integer;
    Height: Integer;
  public
    Bits: Pointer;
    constructor Create(Wnd: HWND; Bpp: Integer=0);
    destructor Destroy; override;
    procedure Draw(const Dist: TCanvas; X, Y: Integer);
    procedure SetSize(Width, Height: Integer);
  end;

  TIcon = class
    Handle: HICON;
    destructor Destroy; override;
  end;

  TCursorImage = HCURSOR;

  Cursors = class
    class function Arrow: TCursorImage;
    class function Cross: TCursorImage;
    class function Ibeam: TCursorImage;
    class function Wait: TCursorImage;
    class function Resource(const Name: WideString): TCursorImage;
  end;

  SysColors = class
    class function Window: Cardinal;
  end;

  function _ToColorRef(Color: Cardinal): COLORREF;

implementation

function _ToColorRef(Color: Cardinal): COLORREF;
begin
  Result := (Color shl 16 and $ff0000) or (Color and $ff00) or (Color shr 16);
end;

{ TFont }

constructor TFont.Create(const Name: WideString; Size: Integer;
    Bold, Italic: Boolean; Quality: TFontQuality=fqDefault);
var
  weight: Integer;
begin
  Self.Name := Name;
  Self.Size := Size;
  Self.Bold := Bold;
  Self.Italic := Italic;
  Self.Quality := Quality;
  if Bold then
    weight := FW_BOLD
  else
    weight := FW_NORMAL;
  Handle := CreateFontW(-Size, 0, 0, 0, weight, DWORD(Italic), 0, 0,
    DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, UINT(Quality),
    DEFAULT_PITCH, PWideChar(Name));
end;

constructor TFont.Create(LogicalFont: TLogicalFont);
var
  ncm: TNonClientMetricsW;
begin
  case LogicalFont of
    lfStatus:
      begin
        ncm.cbSize := SizeOf(ncm);
        SystemParametersInfoW(SPI_GETNONCLIENTMETRICS, SizeOf(ncm), @ncm, 0);
        Handle := CreateFontIndirectW(ncm.lfStatusFont);
      end;
  end;
end;

destructor TFont.Destroy;
begin
  if Handle <> 0 then
    DeleteObject(Handle);
end;


{ TCanvas }

procedure TCanvas.BeginDraw;
begin
  SetBkMode(Handle, TRANSPARENT); // XXX
  OldFont := GetCurrentObject(Handle, OBJ_FONT);
  OldPen := GetCurrentObject(Handle, OBJ_PEN);
end;

procedure TCanvas.EndDraw;
begin
  if GetCurrentObject(Handle, OBJ_FONT) <> OldFont then
    SelectObject(Handle, OldFont);
  if GetCurrentObject(Handle, OBJ_PEN) <> OldPen then
    SelectObject(Handle, OldPen);
  if Brush <> 0 then
    DeleteObject(Brush);
  Brush := 0;
end;

procedure TCanvas.SetTextColor(Color: Cardinal);
begin
  Windows.SetTextColor(Handle, _ToColorRef(Color));
end;

procedure TCanvas.SetTextAlign(Align: TTextAlign);
begin
  Windows.SetTextAlign(Handle, UINT(Align));
end;

procedure TCanvas.SetFont(const Font: TFont);
begin
  if GetCurrentObject(Handle, OBJ_FONT) <> Font.Handle then
    SelectObject(Handle, Font.Handle);
end;

function TCanvas.TextMetrics: TTextMetrics;
var
  tm: TTextMetric;
begin
  GetTextMetrics(Handle, tm);
  with Result do begin
    Height := tm.tmHeight;
    InternalLeading := tm.tmInternalLeading;
  end;
end;

procedure TCanvas.DrawText(X, Y: Integer; const S: string);
begin
  ExtTextOut(Handle, X, Y, 0, nil, PChar(S), Length(S), nil);
end;

procedure TCanvas.DrawText(X, Y: Integer; const S: WideString);
begin
  ExtTextOutW(Handle, X, Y, 0, nil, PWideChar(S), Length(S), nil);
end;

procedure TCanvas.UpdateBrush(Color: Cardinal);
var
  c: COLORREF;
  lb: TLogBrush;
begin
  c := _ToColorRef(Color);
  if Brush = 0 then
    Brush := CreateSolidBrush(c)
  else begin
    GetObject(Brush, SizeOf(lb), @lb);
    if lb.lbColor <> c then begin
      DeleteObject(Brush);
      Brush := CreateSolidBrush(c);
    end;
  end;
end;

procedure TCanvas.DrawRect(X, Y, W, H: Integer; Color: Cardinal);
var
  rc: TRect;
begin
  UpdateBrush(Color);
  rc.Left := X;
  rc.Top := Y;
  rc.Right := X + W;
  rc.Bottom := Y + H;
  Windows.FrameRect(Handle, rc, Brush);
end;

procedure TCanvas.FillRect(X, Y, W, H: Integer; Color: Cardinal);
var
  rc: TRect;
begin
  UpdateBrush(Color);
  rc.Left := X;
  rc.Top := Y;
  rc.Right := X + W;
  rc.Bottom := Y + H;
  Windows.FillRect(Handle, rc, Brush);
end;

procedure TCanvas.UpdatePen(Color: Cardinal);
var
  c: COLORREF;
  lp: TLogPen;
  oldpen: HPEN;
begin
  c := _ToColorRef(Color);
  GetObject(GetCurrentObject(Handle, OBJ_PEN), SizeOf(lp), @lp);
  if lp.lopnColor <> c then begin
    oldpen := SelectObject(Handle, CreatePen(PS_SOLID, 1, c));
    if oldpen <> Self.OldPen then
      DeleteObject(oldpen);
  end;
end;

procedure TCanvas.Line(X1, Y1, X2, Y2: Integer; Color: Cardinal);
var
  pt: array [0..1] of TPoint;
begin
  UpdatePen(Color);
  pt[0].x := X1;
  pt[0].y := Y1;
  pt[1].x := X2;
  pt[1].y := Y2;
  PolyLine(Handle, pt, Length(pt));
end;


{ TMemCanvas }

constructor TMemCanvas.Create(Wnd: HWND; Bpp: Integer=0);
var
  dc: HDC;
begin
  Self.Wnd := Wnd;
  Self.Bpp := Bpp;
  dc := GetDC(Wnd);
  Handle := CreateCompatibleDC(dc);
  ReleaseDC(Wnd, dc);
end;

destructor TMemCanvas.Destroy;
begin
  DeleteObject(Bitmap);
  DeleteDC(Handle);
end;

procedure TMemCanvas.Draw(const Dist: TCanvas; X, Y: Integer);
begin
  BitBlt(Dist.Handle, X, Y, Width, Height, Handle, 0, 0, SRCCOPY);
end;

procedure TMemCanvas.SetSize(Width, Height: Integer);
var
  dc: HDC;
  newbitmap: HBITMAP;
  bi: TBitmapInfo;
begin
  if (Bitmap <> 0) and (Width = Self.Width) and (Height = Self.Height) then
    Exit;
  Self.Width := Width;
  Self.Height := Height;
  dc := GetDC(Wnd);
  if Bpp > 0 then begin
    FillChar(bi, SizeOf(bi), 0);
    bi.bmiHeader.biSize := SizeOf(bi);
    bi.bmiHeader.biWidth := Width;
    bi.bmiHeader.biHeight := Height;
    bi.bmiHeader.biPlanes := 1;
    bi.bmiHeader.biBitCount := Bpp;
    newbitmap := CreateDIBSection(0, bi, DIB_RGB_COLORS, Bits, 0, 0);
  end else
    newbitmap := CreateCompatibleBitmap(dc, Width, Height);
  ReleaseDC(Wnd, dc);
  SelectObject(Handle, newbitmap);
  if Bitmap <> 0 then
    DeleteObject(Bitmap);
  Bitmap := newbitmap;
end;


{ TIcon }

destructor TIcon.Destroy;
begin
  if Handle <> 0 then
    DestroyIcon(Handle);
end;


{ Cursors }

class function Cursors.Arrow: HCURSOR;
begin
  Result := LoadCursorW(0, PWideChar(IDC_ARROW));
end;

class function Cursors.Cross: HCURSOR;
begin
  Result := LoadCursorW(0, PWideChar(IDC_CROSS));
end;

class function Cursors.Ibeam: HCURSOR;
begin
  Result := LoadCursorW(0, PWideChar(IDC_IBEAM));
end;

class function Cursors.Wait: HCURSOR;
begin
  Result := LoadCursorW(0, PWideChar(IDC_WAIT));
end;

class function Cursors.Resource(const Name: WideString): HCURSOR;
begin
  Result := LoadCursorW(HInstance, PWideChar(Name));
end;


{ SysColors }

class function SysColors.Window: Cardinal;
begin
  Result := _ToColorRef(GetSysColor(COLOR_WINDOW));
end;

end.
